import logging,os,json,time,requests
from abstract_utilities import make_list
from datetime import datetime
from ...utils.utils import (
    get_txn_price,
    get_virtual_token_reserves,
    get_virtual_sol_reserves,
    get_token_amount,
    get_sol_amount,
    get_sorted_txn_history,
    get_user_address,
    get_signature,
    get_timestamps,
    get_is_buy,
    if_list_get_single,
    get_mint,
    get_bonding_curve,
    get_string_value
    )
from ...query_functions.call_functions import get_creation_decoded_data,get_pair
def process_transaction_history(txn_history=None,pair_id=None):
    """
    Processes a list of transaction history and returns a cleaned, structured list of transactions
    with Unix timestamps and other essential data.

    Args:
    txn_history (list): List of transaction dictionaries.

    Returns:
    tuple: A tuple containing:
        - processed_data (list): The cleaned list of transactions.
        - init_sol_amount (float): Initial SOL amount.
        - init_token_amount (float): Initial token amount.
        - init_virtualSolReserves (int): Initial virtual Sol reserves.
        - init_virtualTokenReserves (int): Initial virtual token reserves.
        - init_price (float): Initial price derived from virtual reserves.
    """
    processed_data = []
    init_sol_amount = 0
    init_token_amount = 0
    init_virtualSolReserves = 0
    init_virtualTokenReserves = 0
    init_price = 0
    txn_history = get_sorted_txn_history(txn_history)
    if len(txn_history)>0:
        initial_txn = txn_history[0]
    else:
        return None
    pair_id = pair_id or initial_txn.get('pair_id')
    if not pair_id:
        signature = get_signature(txn_history[0])
        pair = get_pair(signature=signature)
        pair = if_list_get_single(pair)
        pair_id = get_string_value(pair,'id')
    pair = get_pair(pair_id = pair_id)
    creator_address = get_user_address(pair)
    if not creator_address:
        pair = get_creation_decoded_data(pair_id = pair_id)
        creator_address = get_user_address(pair)
    # If txn_history is a tuple of lists, flatten it
    mint = get_mint(pair)
    bonding_curve = get_bonding_curve(pair)
    if txn_history:
        # Extract initialization parameters from the first transaction
        init_txn = txn_history[0]
        init_sol_amount = get_sol_amount(init_txn)
        init_token_amount = get_token_amount(init_txn)
        init_virtualSolReserves = get_virtual_sol_reserves(init_txn)
        init_virtualTokenReserves = get_virtual_token_reserves(init_txn)  # Avoid division by zero
        init_price = get_txn_price(init_txn)
    for txn in txn_history:
        try:
            typ = 'Buy' if get_is_buy(txn) else 'Sell'
            sol_amount = get_sol_amount(txn)
            token_amount = get_token_amount(txn)
            price = get_txn_price(txn)
            # Extract and convert timestamp to Unix time
            timestamp = get_timestamps(txn)
            if isinstance(timestamp, str):
                try:
                    # Check for multiple formats of the timestamp
                    if ',' in timestamp:  # RFC 1123 format
                        readable_time = datetime.strptime(timestamp, '%a, %d %b %Y %H:%M:%S %Z')
                    else:  # ISO 8601 format
                        readable_time = datetime.fromisoformat(timestamp)
                    unix_timestamp = int(readable_time.timestamp())
                except Exception as e:
                    logging.error(f"Error converting timestamp from string '{timestamp}': {e}")
                    unix_timestamp = 0
            elif isinstance(timestamp, (int, float)):  # If it's already a numeric timestamp
                unix_timestamp = int(timestamp)
            else:
                logging.error(f"Unsupported timestamp type for value '{timestamp}'")
                unix_timestamp = 0

            processed_data.append({
                'type': typ,
                'timestamp':timestamp,   # Store the Unix timestamp directly
                'price': f"{price:.9f}",
                'sol_amount': f"{sol_amount:.9f}",
                'volume': f"{sol_amount:.9f}",
                'user_address':get_user_address(txn),
                'token_amount': f"{token_amount:.9f}",
                'signature': get_signature(txn),
                'unix_timestamp':unix_timestamp,
                
            })

        except Exception as e:
            logging.error(f"Error processing transaction: {txn}, error: {e}")
            continue
    transaction_history = {
                "processed_data": processed_data,
                "init_sol_amount": init_sol_amount,
                "init_token_amount": init_token_amount,
                "init_virtualSolReserves": init_virtualSolReserves,
                "init_virtualTokenReserves": init_virtualTokenReserves,
                "init_price": init_price,
                "mint":mint,
                "creator_address":creator_address,
                "bonding_curve":bonding_curve
            }
    return transaction_history



from ...utils.utils import get_sorted_txn_history,get_user_address,get_signature,get_sol_amount,get_token_amount
from abstract_utilities import make_list
def tally_profits(txns=None):
    """Organize and calculate profit, volume, and transaction data for each user."""
    txn_history = get_sorted_txn_history(txns)
    txnData_js = {}
    for txn in make_list(txn_history):
        user = get_user_address(txn)
        if isinstance(user,list):
            user = [use for use in user if use]
            if len(user)>0:
                user=user[0]
        for tx in make_list(txn):
            if user and user not in txnData_js:
                txnData_js[user] = {
                    "profits": {"sol": 0, "token": 0},
                    "avgPrice": {'token_amount': 0, 'sol': 0, "avg": 0},
                    "volume": {"sell": 0, "buy": 0, "total": 0},
                    "txns": []
                }
            solAmt = get_sol_amount(tx)
            tknAmt = get_token_amount(tx)
            txnData_js[user]["volume"]["total"] += solAmt

            if tx.get('isbuy'):
                txnData_js[user]["volume"]["buy"] += solAmt
                txnData_js[user]["profits"]["sol"] -= solAmt
            else:
                txnData_js[user]["volume"]["sell"] += solAmt
                txnData_js[user]["profits"]["sol"] += solAmt

            txnData_js[user]["avgPrice"]['token_amount'] += tknAmt
            txnData_js[user]["avgPrice"]['sol'] += solAmt
            # Avoid division by zero:
            token_amt = txnData_js[user]["avgPrice"]['token_amount'] or 1
            txnData_js[user]["avgPrice"]["avg"] = txnData_js[user]["avgPrice"]['sol'] / token_amt
            signature = get_signature(tx)
            txnData_js[user]["txns"].append(signature)
    return txnData_js



from .queries import get_db_from
from .call_pair_data import get_any_pair
def get_metadata_from(searchColumn,searchValue):
    return get_db_from('metadata',searchColumn,searchValue)
def get_meta_data_from_mint(mint):
    return get_metadata_from(searchColumn='mint', searchValue=mint)
def get_meta_data_from_meta_id(meta_id):
    return get_metadata_from(searchColumn='id', searchValue=meta_id)
def get_meta_data_from_pair(pair_data):
    pair = get_any_pair(pair_data)
    meta_id = pair.get('meta_id')
    if not meta_id:
        return []
    return get_meta_data_from_meta_id(meta_id)
#####################################
# Fix #4: get_meta_data uses correct references
#####################################
def get_meta_data(pair_id=None, meta_id=None, mint=None):
    if mint:
        return get_meta_data_from_mint(mint)
    if pair_id:
        return get_meta_data_from_pair(pair_id)
    if meta_id:
        return get_meta_data_from_meta_id(meta_id)
    return []



from ..query_functions.queries import aggregate_rows,query_data
from ..query_functions.fetch_utils import fetch_any_combo,get_db_from,getZipRows
from ...utils import if_list_get_single,get_signature,get_sorted_txn_history



from .queries import get_db_from
def get_logdata_from(columnNames='*',searchColumn=None,searchValue=None):
    return get_db_from('logdata',columnNames=columnNames,searchColumn=searchColumn,searchValue=searchValue)
def get_logdata_from_log_id(log_id):
    return get_logdata_from(searchColumn='id',searchValue=log_id)
def get_signature_from_log_id(log_id):
    return get_logdata_from(columnNames='signature',searchColumn='id',searchValue=log_id)
def get_logdata_from_signature(signature):
    return get_logdata_from(columnNames='*',searchColumn='signature',searchValue=signature)

                    



import logging,os,json,time,requests
import pandas as pd
import mplfinance as mpf
from ...query_functions.call_functions import get_transactions
from ..user_txns.transaction_history import process_transaction_history
from ...utils import convert_chart_data_keys_to_int,get_timestamps,get_sorted_txn_history, get_txn_price,get_sol_amount, get_user_address,convert_timestamp_to_datetime,get_sorted_txn_history
def generate_price_chart(txn_history=None,pair_id=None,save_path=None):
    """
    Generates a candlestick chart with volume indicators from transaction history.

    Parameters:
        txn_history (list of dict): The transaction history data.
        save_path (str): Path to save the generated chart image.

    Returns:
        None
    """
    new_history = []
    if not txn_history and pair_id != None:
        txn_history = get_transactions(pair_id=pair_id)
    elif isinstance(txn_history,int):
        txn_history = get_transactions(pair_id=txn_history)
    save_path = save_path or os.path.join(os.getcwd(),"chart.png")
    txn_history = get_sorted_txn_history(txn_history)
    # Step 1: Prepare transaction data
    for i, txn in enumerate(txn_history):
        try:
            txn['timestamp'] = get_timestamps(txn)
            txn['price'] = get_txn_price(txn)
            txn['volume'] = get_sol_amount(txn)
            txn['user_address'] = get_user_address(txn)
            new_history.append(txn)
        except Exception as e:
            print(f"{e} and txn == {txn}")
    txn_history = new_history
    if not isinstance(txn_history, list) or not all(isinstance(item, dict) for item in txn_history):
        raise ValueError("txn_history must be a list of dictionaries with keys 'timestamp', 'price', and 'volume'.")

    # Step 2: Process and create DataFrame
    processed_txns = process_transaction_history(txn_history)
    processed_txns['save_path']=save_path
    processed_data = processed_txns.get('processed_data')
    try:
        df = pd.DataFrame(processed_data)
    except Exception as e:
        raise ValueError(f"Error creating DataFrame from transaction history: {e}")
    
    if df.empty:
        raise ValueError("DataFrame is empty after processing transaction history.")
    
    # Check for required columns
    required_columns = ['timestamp', 'price', 'volume']
    for col in required_columns:
        if col not in df.columns:
            raise ValueError(f"Missing required column '{col}' in transaction history. Columns found: {df.columns.tolist()}")
    
    # ** Step 3: Ensure numeric columns **
    try:
        df['price'] = pd.to_numeric(df['price'], errors='coerce')
        df['volume'] = pd.to_numeric(df['volume'], errors='coerce')
        
        # Drop any rows where price or volume are NaN
        df.dropna(subset=['price', 'volume'], inplace=True)
        
        if df.empty:
            raise ValueError("DataFrame is empty after converting price and volume to numeric and dropping NaN values.")
    except Exception as e:
        raise ValueError(f"Error converting price and volume columns to numeric: {e}")
    
    # ** Step 4: Convert timestamp to datetime and set as index **
    try:
        df['datetime'] = pd.to_datetime(df['timestamp'], unit='s', errors='coerce', utc=True)        
        if df['datetime'].isnull().all():
            raise ValueError("All timestamps failed to convert. Check 'timestamp' column values.")
        
        # Set the datetime as the index
        df.set_index('datetime', inplace=True)
        df.sort_index(inplace=True)
    except Exception as e:
        raise ValueError(f"Error converting 'timestamp' to 'datetime': {e}")
    
    # ** Step 5: Resample DataFrame into OHLC and Volume for 1-minute intervals **
    try:
        ohlc_dict = {'price': 'ohlc', 'volume': 'sum'}
        df_resampled = df.resample('1min').agg(ohlc_dict).dropna(how='any')
        
        if df_resampled.empty:
            raise ValueError("No data available after resampling. Check if timestamps are too sparse.")
        
        print("Columns after resampling and flattening:", df_resampled.columns)
    except Exception as e:
        raise ValueError(f"Error resampling DataFrame: {e}")
    
    # ** Step 6: Flatten multi-level columns and rename them for mplfinance **
    try:
        df_resampled.columns = ['_'.join(col) if isinstance(col, tuple) else col for col in df_resampled.columns]
        df_resampled.rename(columns={
            'price_open': 'Open', 
            'price_high': 'High', 
            'price_low': 'Low', 
            'price_close': 'Close', 
            'volume_volume': 'Volume'
        }, inplace=True)
        
        #print("Renamed columns after flattening:", df_resampled.columns)
        
        if 'Volume' not in df_resampled.columns:
            raise ValueError(f"'Volume' column is missing. Current columns: {df_resampled.columns}")
    except Exception as e:
        raise ValueError(f"Error renaming DataFrame columns: {e}")
    
    # ** Step 7: Plot the candlestick chart **
    try:
        mc = mpf.make_marketcolors(up='g', down='r', inherit=True)
        s = mpf.make_mpf_style(marketcolors=mc)

        mpf.plot(
            df_resampled,
            type='candle',
            style=s,
            volume=True,
            savefig=dict(fname=save_path, format='png', bbox_inches='tight')
        )
        try:
            chart_data = df_resampled.to_dict()
            
            converted_chart_data_int = convert_chart_data_keys_to_int(chart_data)
            processed_txns['chart_data']=converted_chart_data_int
        except Exception as e:
            print(f"Error plotting candlestick chart: {e}")
    except Exception as e:
        raise ValueError(f"Error plotting candlestick chart: {e}")
    print(f"Chart saved to {save_path}")
    return processed_txns



import logging

# Configure logging
logging.basicConfig(level=logging.INFO)  # Set level to INFO or higher
from psycopg2 import sql, connect
from psycopg2.extras import DictCursor
from abstract_utilities import make_list, SingletonMeta, is_number
from abstract_security import get_env_value
import psycopg2,time

env_path = '/home/solcatcher/.env'

def get_env_val(key):
    return get_env_value(key=key, path=env_path)

RABBITMQ_HOST = get_env_val("SOLCATCHER_AMQP_HOST")
RABBITMQ_PORT = get_env_val("SOLCATCHER_AMQP_PORT")
RABBITMQ_USER = get_env_val("SOLCATCHER_AMQP_USER")
RABBITMQ_NAME = get_env_val("SOLCATCHER_AMQP_NAME")
RABBITMQ_PASSWORD = get_env_val("SOLCATCHER_AMQP_PASSWORD")

POSTGRESQL_HOST = get_env_val("SOLCATCHER_POSTGRESQL_HOST")
POSTGRESQL_PORT = get_env_val("SOLCATCHER_POSTGRESQL_PORT")
POSTGRESQL_USER = get_env_val("SOLCATCHER_POSTGRESQL_USER")
POSTGRESQL_NAME = get_env_val("SOLCATCHER_POSTGRESQL_NAME")
POSTGRESQL_PASSWORD = get_env_val("SOLCATCHER_POSTGRESQL_PASSWORD")

DB_URL = f"postgresql://{POSTGRESQL_USER}:{POSTGRESQL_PASSWORD}@{POSTGRESQL_HOST}:{POSTGRESQL_PORT}/{POSTGRESQL_NAME}"

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

def get_connection():
    """Establish a database connection."""
    return psycopg2.connect(
        dbname=POSTGRESQL_NAME,
        user=POSTGRESQL_USER,
        password=POSTGRESQL_PASSWORD,
        host=POSTGRESQL_HOST,
        port=POSTGRESQL_PORT
    )

def get_all_table_names(schema='public'):
    """Fetch all table names from a specified schema."""
    conn = get_connection()
    query = """
        SELECT table_name
        FROM information_schema.tables
        WHERE table_schema = %s AND table_type = 'BASE TABLE';
    """
    try:
        with conn.cursor() as cursor:
            cursor.execute(query, (schema,))
            table_names = [row[0] for row in cursor.fetchall()]
            return table_names
    except Exception as e:
        print(f"An error occurred in get all tyables: {e}")
    finally:
        conn.close()
def get_time_interval(seconds =0,minutes=0, hours=0, days=0, weeks=0,months=0, years=0,*args,**kwargs):
    """Calculate a UNIX timestamp for the given time offset."""
    secs = 1
    mins = secs*60
    hr = 60 * mins
    day = 24 * hr
    week = 7 * day
    year = 365 * day
    month = year/12
    return ((secs * seconds) + (mins * minutes) + (hr * hours) + (day * days) + (week * weeks) + (month * months) + (year * years))

def get_time(time_interval,*args,**kwargs):
    timeStamp = time.time() - time_interval
    return int(timeStamp)  # Return integer timestamp

def derive_timestamp(timestamp=None,seconds =0,minutes=0, hours=0, days=0, weeks=0,months=0, years=0,*args,**kwargs):
    time_interval = get_time_interval(seconds=seconds,minutes=minutes, hours=hours, days=days, weeks=weeks,months=months, years=years)
    if time_interval:
        timestamp = get_time(time_interval)
    return timestamp
def get_timestamp_from_data(data,*args,**kwargs):
    years = data.get('years',0)
    months = data.get('months',0)
    weeks = data.get('weeks',0)
    days = data.get('days',0)
    hours = data.get('hours',0)
    minutes = data.get('minutes',0)
    seconds = data.get('seconds',0)
    timestamp = data.get('timestamp')
    time_interval = get_time_interval(seconds=seconds,minutes=minutes, hours=hours, days=days, weeks=weeks,months=months, years=years)
    if time_interval:
        timestamp = get_time(time_interval)
    return timestamp

def get_query_result(query, conn):
    """Executes a query and returns the results or commits the transaction."""
    with conn.cursor() as cursor:
        cursor.execute(query)
        if query.strip().lower().startswith("select"):
            return cursor.fetchall()  # Return data for SELECT queries
        conn.commit()

def query_data(query, values=None, error="Error executing query:", zipRows=True):
    """Execute a query and handle transactions with error management."""
    with get_connection() as conn:
        # Choose the cursor type based on whether you want to zip rows with column names
        cursor_factory = DictCursor if zipRows else None
        with conn.cursor(cursor_factory=cursor_factory) as cursor:
            try:
                cursor.execute(query, values)
                result = cursor.fetchall()
                # Log the first row to see its structure
                if result:
                    logging.info("First row data structure: %s", result[0])
                return result
            except Exception as e:
                conn.rollback()
                logging.error("%s %s\nValues: %s\n%s", error, query, values, e)

#####################################
# Fix #1: Correct the usage in fetch_any_combo
#####################################
def fetch_any_combo(columnNames='*',
                    tableName=None,
                    searchColumn=None,
                    searchValue=None,
                    anyValue=False,
                    zipIt=True,
                    schema='public'):
    """
    Fetch rows based on dynamic SQL built from parameters.
    
    :param columnNames: Comma separated columns or '*' for all.
    :param tableName: The table to query. Must not be None or '*'.
    :param searchColumn: The column on which to filter.
    :param searchValue: The value to match in searchColumn.
    :param anyValue: If True, uses = ANY(%s) for arrays.
    :param zipit: If True, uses DictCursor in query_data.
    :param schema: The DB schema.
    """
    if not tableName or tableName == '*':
        logging.error("Invalid tableName provided to fetch_any_combo: %s", tableName)
        return []  # or raise an Exception

    # Build the SELECT list
    if columnNames == '*':
        select_cols = sql.SQL('*')
    else:
        # Convert "col1,col2" -> [col1, col2]
        col_list = [c.strip() for c in columnNames.split(',')]
        select_cols = sql.SQL(", ").join(sql.Identifier(col) for col in col_list)

    # Build the base query: SELECT ... FROM schema.tableName
    base_query = sql.SQL("SELECT {} FROM {}.{}").format(
        select_cols,
        sql.Identifier(schema),
        sql.Identifier(tableName)
    )

    # Build the WHERE clause if needed
    params = []
    if searchColumn and searchValue is not None:
        if anyValue:
            base_query += sql.SQL(" WHERE {} = ANY(%s)").format(sql.Identifier(searchColumn))
            params.append(searchValue if isinstance(searchValue, list) else [searchValue])
        else:
            base_query += sql.SQL(" WHERE {} = %s").format(sql.Identifier(searchColumn))
            params.append(searchValue)

    result = query_data(base_query, values=params, zipRows=zipit)
    if zipIt:
        result = getZipRows(tableName, result)
    return result
def get_anything(*args, **kwargs):
    if args:
        for arg in args:
            if 'tableName' not in kwargs:
                kwargs['tableName'] = arg
    response = fetch_any_combo(**kwargs)
    logging.info("Received data: %s", response)  # Log to see the data
    if isinstance(response, list) and len(response) == 1:
        response = response[0]
    return response

def get_table_name_from_query(query):
    """Extract table name from SQL query."""
    if isinstance(query, sql.Composed):
        query = query.as_string(get_connection())  # Convert to string
    parts = query.split()
    if 'from' in parts:
        return parts[parts.index('from') + 1]
    return None

def get_all_pair_info(**kwargs):
    pair = get_pair(**kwargs)
    if not pair:
        return {}
    pair = pair[0]
    pair['meta_data'] = get_meta_data_from_meta_id(pair.get('meta_id'))
    pair['txns'] = get_all_txns_for_pair_id(pair.get('id'))
    return pair

class columnNamesManager(metaclass=SingletonMeta):
    def __init__(self):
        if not hasattr(self, 'initialized'):
            self.initialized = True
            self.columnNames = {}

    def get_column_names(self, tableName, schema='public'):
        if tableName not in self.columnNames:
            self.columnNames[tableName] = self.fetch_column_names(tableName, schema)
        return self.columnNames[tableName]

    def fetch_column_names(self, tableName, schema='public'):
        query = """
            SELECT column_name
            FROM information_schema.columns
            WHERE table_name = %s AND table_schema = %s
            ORDER BY ordinal_position;
        """
        results = query_data(query, [tableName, schema], error='Error fetching column names')
        if not results:
            logging.warning(f"No columns found for table {tableName} in schema {schema}")
            return []
        return [row[0] for row in results]

    def zip_rows(self, tableName, rows, schema='public'):
        column_names = self.get_column_names(tableName, schema)
        return [dict(zip(column_names, make_list(row))) for row in rows]

def get_pair_id(*args, **kwargs):
    """
    Example function that tries to retrieve a pair's ID by various criteria.
    There's a potential bug with 'is_mint(obj)' references, so watch out for that.
    """
    id_val = None
    for key,value in kwargs.items():
        if key == 'pair_id':
            id_val = value
        if key == 'pair' and isinstance(value, dict):
            id_val = value.get('id')
        if key == 'mint':
            rows = fetch_any_combo(tableName='transactions', searchColumn='mint', searchValue=value)
            if rows:
                id_val = rows[0].get('id')
        if key == 'signature':
            rows = fetch_any_combo(tableName='transactions', searchColumn='signature', searchValue=value)
            if rows:
                id_val = rows[0].get('id')

    for arg in args:
        if is_number(arg):
            id_val = arg
        elif isinstance(arg, dict):
            id_val = arg.get('id')
        # Potential bug fix #2: changed 'obj' -> 'arg'
        # elif is_signature(arg):
        #     ...
        # elif is_mint(arg):
        #     ...
        # etc.

    if id_val:
        return id_val
def get_max_for_row(tableName):
    return query_data(f"SELECT COUNT(*) FROM {tableName};")
#####################################
# Fix #2: Correctly use get_pair_from_* in get_pair
#####################################
def get_pair(mint=None, signature=None, pair_id=None):
    """Retrieve row(s) from 'pairs' table by mint, signature, or pair_id."""
    if pair_id:
        return get_pair_from_pair_id(pair_id)
    elif mint:
        return get_pair_from_mint(mint)
    elif signature:
        return get_pair_from_signature(signature)
    return []

def get_pair_from_mint(mint):
    response = fetch_any_combo(tableName='pairs', searchColumn="mint", searchValue=mint, zipit=True)
    return getZipRows('pairs', response)

def get_pair_from_signature(signature):
    response = fetch_any_combo(tableName='pairs', searchColumn="signature", searchValue=signature, zipit=True)
    return getZipRows('pairs', response)

def get_pair_from_pair_id(pair_id):
    response = fetch_any_combo(tableName='pairs', searchColumn="id", searchValue=pair_id, zipit=True)
    return getZipRows('pairs', response)

def getZipRows(tableName, rows, schema='public'):
    column_names = get_column_names(tableName, schema)
    return [dict(zip(column_names, row)) for row in make_list(rows) if row]

def get_column_names(tableName, schema='public'):
    return columnNamesManager().get_column_names(tableName, schema)

def get_transaction_from_txn_id(txn_id):
    response = fetch_any_combo(tableName='transactions', searchColumn="id", searchValue=txn_id, zipit=True)
    return getZipRows('transactions', response)
def get_transaction_from_signature(signature):
    response = fetch_any_combo(tableName='transactions', searchColumn="signature", searchValue=signature, zipit=True)
    return getZipRows('transactions', response)

def get_transaction_from_log_id(log_id):
    """
    Example that first fetches signature from logdata, then fetches transactions by signature.
    """
    signature_rows = fetch_any_combo(columnNames='signature',
                                     tableName='logdata',
                                     searchColumn='id',
                                     searchValue=log_id,
                                     zipit=True)
    if signature_rows and len(signature_rows) == 1:
        signature = signature_rows[0].get('signature')
        tx_rows = fetch_any_combo(tableName='transactions',
                                  searchColumn='signature',
                                  searchValue=signature,
                                  zipit=True)
        return getZipRows('transactions', tx_rows)
    return []

def get_all_txns_for_pair_id(pair_id):
    response = fetch_any_combo(tableName='transactions', searchColumn='pair_id', searchValue=pair_id, zipit=True)
    return getZipRows('transactions', response)

#####################################
# Fix #3: Add missing signature clause in get_transactions
#####################################
def get_transactions(txn_id=None, log_id=None, signature=None, pair_id=None):
    if pair_id:
        return get_all_txns_for_pair_id(pair_id)
    if txn_id:
        return get_transaction_from_txn_id(txn_id)
    if log_id:
        return get_transaction_from_log_id(log_id)
    if signature:
        return get_transaction_from_signature(signature)
    return []

def get_meta_data_from_mint(mint):
    response = fetch_any_combo(tableName='metadata', searchColumn='mint', searchValue=mint, zipit=True)
    return getZipRows('metadata', response)

def get_meta_data_from_meta_id(meta_id):
    response = fetch_any_combo(tableName='metadata', searchColumn='id', searchValue=meta_id, zipit=True)
    return getZipRows('metadata', response)

def get_meta_data_from_pair_id(pair_id):
    pair_rows = get_pair_from_pair_id(pair_id)
    if not pair_rows:
        return []
    pair = pair_rows[0]
    meta_id = pair.get('meta_id')
    if not meta_id:
        return []
    return get_meta_data_from_meta_id(meta_id)

#####################################
# Fix #4: get_meta_data uses correct references
#####################################
def get_meta_data(pair_id=None, meta_id=None, mint=None):
    if mint:
        return get_meta_data_from_mint(mint)
    if pair_id:
        return get_meta_data_from_pair_id(pair_id)
    if meta_id:
        return get_meta_data_from_meta_id(meta_id)
    return []

#####################################
# Fix #5: Ensure get_all_pair_data aligns with the fixed get_pair
#####################################
def get_all_pair_data(mint=None, signature=None, pair_id=None):
    """
    Gathers pair info, associated transactions, and metadata.
    """
    pair_rows = get_pair(mint=mint, signature=signature, pair_id=pair_id)
    if not pair_rows:
        return {}
    pair = pair_rows[0]
    pair_id_val = pair.get('id')
    mint_val = pair.get('mint')

    pair_data = {}
    pair_data["pair"] = pair
    pair_data["txns"] = get_transactions(pair_id=pair_id_val)
    pair_data["metadata"] = get_meta_data(mint=mint_val)
    return pair_data

def fetch_filtered_transactions_paginated(
    sol_amount, 
    operator=">",
    timestamp_operator=None, 
    limit=50, 
    offset=0,
    years=0,
    months=0,
    weeks=0,
    days=0,
    hours=0,
    minutes=0,
    seconds=0,
    timestamp=None,
    *args,
    **kwargs
):
    # Build the base SQL query.
    # Note that we're filtering on the first element of the JSON array in t.tcns
    query = f"""
        SELECT
            p.*, 
            t.*,
            m.*,
            t2.id AS related_txn
        FROM 
            transactions t
        JOIN 
            pairs p ON t.pair_id = p.id
        LEFT JOIN 
            metadata m ON p.meta_id = m.id
        LEFT JOIN 
            transactions t2 ON t.pair_id = t2.pair_id AND t.signature != t2.signature
        WHERE 
            t.signature IN (
                SELECT signature FROM pairs WHERE signature IS NOT NULL
            )
        AND 
            t.program_id = p.program_id
        AND 
            EXISTS (
                SELECT 1
                FROM jsonb_array_elements(t.tcns) AS elem
                WHERE (elem ->> 'sol_amount')::numeric {operator} %s
    """
    params = [sol_amount]

    # Derive a timestamp cutoff if provided (assuming derive_timestamp is implemented)
    ts = derive_timestamp(
            years=years,
            months=months,
            weeks=weeks,
            days=days,
            hours=hours,
            minutes=minutes,
            seconds=seconds,
            timestamp=timestamp)
    
    if ts and timestamp_operator:
        query += f" AND (elem ->> 'timestamp')::bigint {timestamp_operator} %s"
        params.append(int(ts))
    
    # Close the EXISTS clause and add ordering/pagination
    query += ") ORDER BY t.updated_at DESC LIMIT %s OFFSET %s;"
    params.extend([limit, offset])

    return query_data(query, params)
